from unittest import mock
import asyncio

import pytest

from broqer.coro_queue import CoroQueue, AsyncMode
from broqer import NONE


@pytest.mark.parametrize('mode', AsyncMode)
@pytest.mark.parametrize('args', [(), (None,), (1, 2, 3)])
@pytest.mark.asyncio
async def test_simple_run(mode, args):
    excpected_result = args

    async def _coro(*args):
        await asyncio.sleep(0)
        return args

    coro_queue = CoroQueue(coro=_coro, mode=mode)

    # test the first call and wait for finish
    future = coro_queue.schedule(*args)
    assert (await future) == excpected_result

    if mode == AsyncMode.LAST_DISTINCT:
        # for LAST_DISTINCT, the following results will be broqer.NONE
        excpected_result = NONE

    # test a second time
    future = coro_queue.schedule(*args)
    assert (await future) == excpected_result

    await asyncio.sleep(0.001)

    # test a third time
    future = coro_queue.schedule(*args)
    assert (await future) == excpected_result


@pytest.mark.parametrize('mode', AsyncMode)
@pytest.mark.asyncio
async def test_exception(mode):
    async def _coro(fail):
        if fail:
            raise ZeroDivisionError()

    coro_queue = CoroQueue(coro=_coro, mode=mode)

    # test the first call and wait for finish
    future = coro_queue.schedule(True)
    with pytest.raises(ZeroDivisionError):
        await future

    # test a second time
    future = coro_queue.schedule(False)
    assert (await future) == None

    # test a third time
    future = coro_queue.schedule(True)
    with pytest.raises(ZeroDivisionError):
        await future


@pytest.mark.asyncio
async def test_concurrent():
    callback = mock.Mock()
    event = asyncio.Event()

    async def _coro(index):
        callback(f'Start {index}')
        await event.wait()
        callback(f'End {index}')
        return index

    coro_queue = CoroQueue(coro=_coro, mode=AsyncMode.CONCURRENT)

    # make two concurrent calls
    result1 = coro_queue.schedule(1)
    result2 = coro_queue.schedule(2)

    await asyncio.sleep(0)

    callback.assert_has_calls([mock.call('Start 1'), mock.call('Start 2')])
    callback.reset_mock()

    event.set()
    await asyncio.sleep(0.001)
    callback.assert_has_calls([mock.call('End 1'), mock.call('End 2')], any_order=True)

    assert result1.result() == 1
    assert result2.result() == 2

    # make a non-concurrent call
    event.set()
    assert (await coro_queue.schedule(3)) == 3


@pytest.mark.parametrize('wait', [True, False])
@pytest.mark.asyncio
async def test_interrupt(wait):
    callback = mock.Mock()
    event = asyncio.Event()

    async def _coro(index):
        callback(f'Start {index}')
        await event.wait()
        callback(f'End {index}')
        return index

    coro_queue = CoroQueue(coro=_coro, mode=AsyncMode.INTERRUPT)

    # make two concurrent calls, the first one will be interrupted (canceld)
    result1 = coro_queue.schedule(1)
    if wait:
        await asyncio.sleep(0.0001)
    result2 = coro_queue.schedule(2)

    await asyncio.sleep(0)

    callback.assert_called_with('Start 2')
    callback.reset_mock()

    event.set()
    await asyncio.sleep(0.001)
    callback.assert_called_once_with('End 2')

    assert result1.result() == NONE
    assert result2.result() == 2

    # make a non-concurrent call
    event.set()
    assert (await coro_queue.schedule(3)) == 3


@pytest.mark.asyncio
async def test_queue():
    callback = mock.Mock()
    event = asyncio.Event()

    async def _coro(index):
        callback(f'Start {index}')
        await event.wait()
        callback(f'End {index}')
        return index

    coro_queue = CoroQueue(coro=_coro, mode=AsyncMode.QUEUE)

    # make two concurrent calls
    result1 = coro_queue.schedule(1)
    result2 = coro_queue.schedule(2)

    await asyncio.sleep(0.001)
    event.set()
    event.clear()

    await asyncio.sleep(0.001)

    callback.assert_has_calls([mock.call('Start 1'), mock.call('End 1'), mock.call('Start 2')])

    assert result1.result() == 1
    assert not result2.done()

    callback.reset_mock()

    event.set()
    event.clear()
    await asyncio.sleep(0.001)

    callback.assert_called_once_with('End 2')

    # make a non-concurrent call
    event.set()
    assert (await coro_queue.schedule(3)) == 3


@pytest.mark.parametrize('distinct', [True, False])
@pytest.mark.asyncio
async def test_last(distinct):
    callback = mock.Mock()
    event = asyncio.Event()

    async def _coro(index):
        callback(f'Start {index}')
        await event.wait()
        callback(f'End {index}')
        return index

    mode = AsyncMode.LAST_DISTINCT if distinct else AsyncMode.LAST
    coro_queue = CoroQueue(coro=_coro, mode=mode)

    # make three concurrent calls
    result1 = coro_queue.schedule(1)
    result2 = coro_queue.schedule(2)
    result3 = coro_queue.schedule(1)

    await asyncio.sleep(0.001)
    event.set()
    await asyncio.sleep(0.001)

    expected_calls = [mock.call('Start 1'), mock.call('End 1')]

    if not distinct:
        expected_calls += [mock.call('Start 1'), mock.call('End 1')]

    callback.assert_has_calls(expected_calls)

    assert result1.result() == 1
    assert result2.result() == NONE
    assert result3.result() == (NONE if distinct else 1)

    # make a non-concurrent call
    event.set()
    assert (await coro_queue.schedule(3)) == 3


@pytest.mark.parametrize('wait', [True, False])
@pytest.mark.asyncio
async def test_skip(wait):
    callback = mock.Mock()
    event = asyncio.Event()

    async def _coro(index):
        callback(f'Start {index}')
        await event.wait()
        callback(f'End {index}')
        return index

    coro_queue = CoroQueue(coro=_coro, mode=AsyncMode.SKIP)

    # make two concurrent calls, the first one will be interrupted (canceld)
    result1 = coro_queue.schedule(1)
    if wait:
        await asyncio.sleep(0.0001)
    result2 = coro_queue.schedule(2)

    await asyncio.sleep(0)

    callback.assert_called_with('Start 1')
    callback.reset_mock()

    event.set()
    await asyncio.sleep(0.001)
    callback.assert_called_once_with('End 1')

    assert result1.result() == 1
    assert result2.result() == NONE

    # make a non-concurrent call
    event.set()
    assert (await coro_queue.schedule(3)) == 3


"""
Apply ``coro`` to each emitted value allowing async processing

Usage:

>>> import asyncio
>>> from broqer import Value, Sink, op
>>> s = Value()

>>> async def delay_add(a):
...     print('Starting with argument', a)
...     await asyncio.sleep(0.015)
...     result = a + 1
...     print('Finished with argument', a)
...     return result

AsyncMode: CONCURRENT (is default)

>>> s.emit(0)
>>> _d = (s | op.MapAsync(delay_add)).subscribe(Sink())
>>> s.emit(1)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.02))
Starting with argument 0
Starting with argument 1
Finished with argument 0
Finished with argument 1
>>> _d.dispose()

AsyncMode: INTERRUPT

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.INTERRUPT))
>>> _d = o.subscribe(Sink(print))
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.005))
Starting with argument 0
>>> s.emit(1)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.02))
Starting with argument 1
Finished with argument 1
2
>>> _d.dispose()

AsyncMode: QUEUE

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.QUEUE))
>>> _d = o.subscribe(Sink(print))
>>> s.emit(1)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.04))
Starting with argument 0
Finished with argument 0
1
Starting with argument 1
Finished with argument 1
2
>>> _d.dispose()

AsyncMode: LAST

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.LAST))
>>> _d = o.subscribe(Sink(print))
>>> s.emit(1)
>>> s.emit(2)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.04))
Starting with argument 0
Finished with argument 0
1
Starting with argument 2
Finished with argument 2
3
>>> _d.dispose()

AsyncMode: SKIP

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.SKIP))
>>> _d = o.subscribe(Sink(print))
>>> s.emit(1)
>>> s.emit(2)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.04))
Starting with argument 0
Finished with argument 0
1
>>> _d.dispose()

Using error_callback:

>>> def cb(*e):
...     print('Got error')

>>> s.emit('abc')
>>> _d = (s | op.MapAsync(delay_add, error_callback=cb)).subscribe(Sink(print))
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.02))
Starting with argument abc
Got error
>>> _d.dispose()
"""
import asyncio
import sys
from functools import wraps
from typing import Any  # noqa: F401

# pylint: disable=cyclic-import
from broqer.operator import Operator
from broqer.coro_queue import CoroQueue, AsyncMode, wrap_coro
from broqer import Publisher, default_error_handler, NONE


class MapAsync(Operator):  # pylint: disable=too-many-instance-attributes
    """ Apply ``coro(*args, value, **kwargs)`` to each emitted value allow
    async processing.

    :param coro: coroutine to be applied on emit
    :param \\*args: variable arguments to be used for calling coro
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked as (\\*value)
    :param \\*\\*kwargs: keyword arguments to be used for calling coro

    :ivar scheduled: Publisher emitting the value when coroutine is actually
        started.
    """
    def __init__(self,
                 coro, *args, mode=AsyncMode.CONCURRENT,
                 error_callback=default_error_handler, unpack: bool = False,
                 **kwargs
                 ) -> None:
        Operator.__init__(self)
        _coro = wrap_coro(coro, unpack, *args, **kwargs)
        self._coro_queue = CoroQueue(_coro, mode=mode)
        self._error_callback = error_callback

    def emit(self, value: Any, who: Publisher) -> None:
        if who is not self._originator:
            raise ValueError('Emit from non assigned publisher')

        future = self._coro_queue.schedule(value)
        future.add_done_callback(self._done)

    def _done(self, future: asyncio.Future):
        try:
            result = future.result()
        except Exception:  # pylint: disable=broad-except
            self._error_callback(*sys.exc_info())
        else:
            if result != NONE:
                Publisher.notify(self, result)


def build_map_async(coro=None, *,
                    mode: AsyncMode = AsyncMode.CONCURRENT,
                    error_callback=default_error_handler,
                    unpack: bool = False):
    """ Decorator to wrap a function to return a Map operator.

    :param coro: coroutine to be wrapped
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_map_async(coro):
        return MapAsync(coro, mode=mode, error_callback=error_callback,
                        unpack=unpack)

    if coro:
        return _build_map_async(coro)

    return _build_map_async


def build_map_async_factory(coro=None, *,
                            mode: AsyncMode = AsyncMode.CONCURRENT,
                            error_callback=default_error_handler,
                            unpack: bool = False):
    """ Decorator to wrap a coroutine to return a factory for MapAsync
        operators.

    :param coro: coroutine to be wrapped
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked (*value)
    """
    _mode = mode

    def _build_map_async(coro):
        @wraps(coro)
        def _wrapper(*args, mode=None, **kwargs) -> MapAsync:
            if ('unpack' in kwargs) or ('error_callback' in kwargs):
                raise TypeError('"unpack" and "error_callback" has to '
                                'be defined by decorator')
            if mode is None:
                mode = _mode
            return MapAsync(coro, *args, mode=mode, unpack=unpack,
                            error_callback=error_callback, **kwargs)
        return _wrapper

    if coro:
        return _build_map_async(coro)

    return _build_map_async



""" CoroQueue handles running a coroutine depending on a given mode
"""
import asyncio
from collections import deque
from enum import Enum
from typing import Any, Deque, Optional, Tuple  # noqa: F401
from functools import partial

from broqer import NONE


def wrap_coro(coro, unpack, *args, **kwargs):
    """ building a coroutine receiving one argument and call it curried
    with *args and **kwargs and unpack it (if unpack is set)
    """
    if unpack:
        async def _coro(value):
            return await coro(*args, *value, **kwargs)
    else:
        async def _coro(value):
            return await coro(*args, value, **kwargs)

    return _coro


class AsyncMode(Enum):
    """ AyncMode defines how to act when an emit happens while an scheduled
    coroutine is running """
    CONCURRENT = 1  # just run coroutines concurrent
    INTERRUPT = 2  # cancel running and call for new value
    QUEUE = 3  # queue the value(s) and call after coroutine is finished
    LAST = 4  # use last emitted value after coroutine is finished
    LAST_DISTINCT = 5  # like LAST but only when value has changed
    SKIP = 6  # skip values emitted during coroutine is running


class CoroQueue:  # pylint: disable=too-few-public-methods
    """ Schedules the running of a coroutine given on a mode
    :param coro: Coroutine to be scheduled
    :param mode: scheduling mode (see AsyncMode)
    """
    def __init__(self, coro, mode=AsyncMode.CONCURRENT):
        self._coro = coro
        self._mode = mode

        # ._last_args is used for LAST_DISTINCT and keeps the last arguments
        self._last_args = None  # type: Optional[Tuple]

        # ._task is the reference to a running coroutine encapsulated as task
        self._task = None  # type: Optional[asyncio.Future]

        # queue is initialized with following sizes:
        # Mode:                size:
        # QUEUE                unlimited
        # LAST, LAST_DISTINCT  1
        # all others           no queue used
        self._queue = \
            None  # type: Optional[Deque[Tuple[Tuple, asyncio.Future]]]

        if mode in (AsyncMode.QUEUE, AsyncMode.LAST, AsyncMode.LAST_DISTINCT):
            maxlen = (None if mode is AsyncMode.QUEUE else 1)
            self._queue = deque(maxlen=maxlen)

    def schedule(self, *args: Any) -> asyncio.Future:
        """ Schedule a coroutine run with the given arguments
        :param *args: variable length arguments
        """
        future = asyncio.Future()  # type: asyncio.Future

        # check if a coroutine is already running
        if self._task is not None:
            # append to queue if a queue is used in this mode
            if self._queue is not None:
                if self._queue.maxlen == 1 and len(self._queue) == 1:
                    _, queued_future = self._queue.popleft()
                    queued_future.set_result(NONE)

                self._queue.append((args, future))
                return future

            # in SKIP mode just do nothin with this emit
            if self._mode is AsyncMode.SKIP:
                future.set_result(NONE)
                return future

            # cancel the task if INTERRUPT mode is used
            if self._mode is AsyncMode.INTERRUPT and not self._task.done():
                self._task.cancel()

        # start the coroutine
        self._start_task(args, future)

        return future

    def _start_task(self, args: Tuple, future: asyncio.Future):
        """ Start the coroutine as task """

        # when LAST_DISTINCT is used only start coroutine when value changed
        if self._mode is AsyncMode.LAST_DISTINCT and args == self._last_args:
            self._task = None
            future.set_result(NONE)
            return

        # store the value to be emitted for LAST_DISTINCT
        self._last_args = args

        # create a task out of it and add ._task_done as callback
        self._task = asyncio.ensure_future(self._coro(*args))
        self._task.add_done_callback(partial(self._handle_done, future))

    def _handle_done(self, result_future: asyncio.Future, task: asyncio.Task):
        try:
            result = task.result()
        except asyncio.CancelledError:  # happend in INTERRUPT mode
            result_future.set_result(NONE)
        except Exception as exception:  # pylint: disable=broad-except
            result_future.set_exception(exception)
        else:
            result_future.set_result(result)

        if self._queue:
            args, future = self._queue.popleft()

            # start the coroutine
            self._start_task(args, future)
        else:
            self._task = None



""" Implementing Value """

from typing import Any

# pylint: disable=cyclic-import
from broqer import Publisher, NONE
from broqer.operator import Operator


class Value(Operator):
    """
    Value is a publisher and subscriber.

    >>> from broqer import Sink

    >>> s = Value(0)
    >>> _d = s.subscribe(Sink(print))
    0
    >>> s.emit(1)
    1
    """
    def __init__(self, init=NONE):
        Operator.__init__(self)
        self._state = init

    def emit(self, value: Any,
             who: Publisher = None) -> None:  # pylint: disable=unused-argument
        if self._originator is not None and self._originator is not who:
            raise ValueError('Emit from non assigned publisher')

        return Publisher.notify(self, value)

    notify = Publisher.notify


def dependent_subscribe(publisher: Publisher, value: Value):
    """ Let `value` subscribe to `publisher` only when `value` itself is
        subscribed
    :param publisher: publisher to be subscribed, when value is subscribed
    :param value: value, which will receive .emit calls from `publisher`
    """
    def _on_subscription(existing_subscription: bool):
        if existing_subscription:
            publisher.subscribe(value)
        else:
            publisher.unsubscribe(value)

    value.register_on_subscription_callback(_on_subscription)
